/* Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.
 *
 * This software and the information contained herein is PROPRIETARY and
 * CONFIDENTIAL to NVIDIA and is being provided under the terms and
 * conditions of a Non-Disclosure Agreement.  Any reproduction or
 * disclosure to any third party without the express written consent of
 * NVIDIA is prohibited.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.  This source code is a "commercial item" as
 * that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer software" and "commercial computer software
 * documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <GL/glew.h>
#include <GL/glut.h>

#define NV_REPORT_COMPILE_ERRORS    // Defined to print compiler errors.

#include "nvSDKPath.h"
#include "nvShaderUtils.h"
#include "nvGlutManipulators.h"
#include "nvGlutWidgets.h"
#include "nvModel.h"

// This example shows how to use bindable uniform buffers and draw instanced in order to 
// render tessellated curved surfaces. The vertices and the indices are stored in buffer 
// objects and bound to the shader. Patches are drawn by rendering a single tessellated
// quad multiple times, using instancing to do this in a single call and the instance id
// to index the index buffer. Indices are stored as integers using the integer support
// in gpu shader4.
//
// This is idea is based on this paper:
// http://www.labri.fr/publications/is/2005/BS05/


#define USE_GL_TIMER
//#define USE_STRIPS

// Vertex shader.
static const char * s_tessellationVertexShaderText =
    "#version 120\n"
    "#extension GL_EXT_bindable_uniform : enable\n"
    "#extension GL_EXT_gpu_shader4 : enable\n"
    "\n"
    "bindable uniform vec4 vertices[%d];\n"
    "\n"
    "varying vec3 v_Normal;\n"
    "\n"
    "void main(void)\n"
    "{\n"
    "    vec3 p[16];\n"
    "    \n"
    "    for (int i = 0; i < 16; i++) {\n"
    "        p[i] = vertices[gl_InstanceID * 16 + i].xyz;\n"
    "    }\n"
    "    \n"
    "    vec2 uv = gl_Vertex.xy;\n"
    "    \n"
    "    vec2 B0 = (1 - uv) * (1 - uv) * (1 - uv);\n"
    "    vec2 B1 = 3 * uv * (1 - uv) * (1 - uv);\n"
    "    vec2 B2 = 3 * uv * uv * (1 - uv);\n"
    "    vec2 B3 = uv * uv * uv;\n"
    "    \n"
    "    vec3 pos = \n"
    "        (B0.x * p[ 0] + B1.x * p[ 1] + B2.x * p[ 2] + B3.x * p[ 3]) * B0.y +\n"
    "        (B0.x * p[ 4] + B1.x * p[ 5] + B2.x * p[ 6] + B3.x * p[ 7]) * B1.y +\n"
    "        (B0.x * p[ 8] + B1.x * p[ 9] + B2.x * p[10] + B3.x * p[11]) * B2.y +\n"
    "        (B0.x * p[12] + B1.x * p[13] + B2.x * p[14] + B3.x * p[15]) * B3.y;\n"
    "    \n"
    "    vec2 T0 = (1 - uv) * (1 - uv);\n"
    "    vec2 T1 = 2 * uv * (1 - uv);\n"
    "    vec2 T2 = uv * uv;\n"
    "    \n"
    "    vec3 dv = \n"
    "        (B0.x * (p[ 4]-p[0]) + B1.x * (p[ 5]-p[1]) + B2.x * (p[ 6]-p[ 2]) + B3.x * (p[ 7]-p[ 3])) * T0.y +\n"
    "        (B0.x * (p[ 8]-p[4]) + B1.x * (p[ 9]-p[5]) + B2.x * (p[10]-p[ 6]) + B3.x * (p[11]-p[ 7])) * T1.y +\n"
    "        (B0.x * (p[12]-p[8]) + B1.x * (p[13]-p[9]) + B2.x * (p[14]-p[10]) + B3.x * (p[15]-p[11])) * T2.y;\n"
    "    \n"
    "    vec3 du = \n"
    "        (T0.x * (p[ 1]-p[ 0]) + T1.x * (p[ 2]-p[ 1]) + T2.x * (p[ 3]-p[ 2])) * B0.y +\n"
    "        (T0.x * (p[ 5]-p[ 4]) + T1.x * (p[ 6]-p[ 5]) + T2.x * (p[ 7]-p[ 6])) * B1.y +\n"
    "        (T0.x * (p[ 9]-p[ 8]) + T1.x * (p[10]-p[ 9]) + T2.x * (p[11]-p[10])) * B2.y +\n"
    "        (T0.x * (p[13]-p[12]) + T1.x * (p[14]-p[13]) + T2.x * (p[15]-p[14])) * B3.y;\n"
    "    \n"
    "    gl_Position = gl_ModelViewProjectionMatrix * vec4(pos, 1.0);\n"
//	"    gl_TexCoord[0].xy = uv;\n"
    "    vec3 nor = cross(du, dv);\n"
    "    nor = (length(nor) != 0) ? normalize(nor) : vec3(0.0);\n"
    "    v_Normal = gl_NormalMatrix * nor;\n"
    "}\n"
    "\n";

static const char * s_standardVertexShaderText =
    "#version 120\n"
    "\n"
    "varying vec3 v_Normal;\n"
    "\n"
    "void main(void)\n"
    "{\n"
	"    gl_Position = ftransform();\n"
//	"    gl_TexCoord[0] = gl_MultiTexCoord0;\n"
    "    v_Normal = gl_NormalMatrix * gl_Normal;\n"
    "}\n"
    "\n";

// Default fragment shader.
static const char * s_fragmentShaderText =
    "varying vec3 v_Normal;\n"
    "\n"
    "uniform vec3 color;\n"
	"uniform sampler2D tex;\n"
    "\n"
    "void main() {\n"
//	"    vec3 c = texture2D(tex,gl_TexCoord[0].st).rgb;\n"
	"    \n"
	"    vec3 N = normalize(v_Normal);\n"
	"    vec3 L = normalize(gl_LightSource[0].position.xyz);\n"
	"    vec3 H = normalize(gl_LightSource[0].halfVector.xyz);\n"
	"    \n"
	"    vec3 ambient = color * 0.1;\n"
	"    vec3 diffuse = color * 0.9 * max(dot(L, N), 0.0);\n"
	"    vec3 specular = vec3(1.0, 1.0, 1.0) * pow(max(dot(H, N), 0.0), 16.0);\n"
	"    \n"
    "    gl_FragColor = vec4(ambient + diffuse + specular, 1.0);\n"
    //"    gl_FragColor = vec4(color, 1.0);\n"
    "}\n";

// Window properties.
static int s_width = 512;
static int s_height = 512;

// Manipulators.
static nv::GlutExamine s_trackball;

// Handles.
static GLuint s_tessellationProgram = 0;
static GLuint s_standardProgram = 0;
static GLuint s_vbHandle = 0;
static GLuint s_ibHandle = 0;

// Geometry.
#include "teapot.h"

static int s_patchVertexCount;
static int s_patchIndexCount;
const static int c_teapotVertexCount = sizeof(teapotVertices) / (3 * sizeof(float));
const static int c_teapotIndexCount = sizeof(teapotIndices) / sizeof(int);
const static int c_teapotPatchCount = c_teapotIndexCount / 16;

static nv::Model model;
static GLuint s_indexBuffer;
static GLuint s_indexCount;
static GLuint s_vertexBuffer;

// UI
static nv::GlutUIContext ui;

enum DrawMethod
{
	DrawMethod_Tessellation,
	DrawMethod_Evaluators,
	DrawMethod_Static,
};

// Rendering options.
static bool s_drawWireframe = false;
static bool s_drawControlMesh = false;
static int s_drawMethod = DrawMethod_Tessellation;
static bool s_animate = true;
static bool s_preWarmCache = true;
static int s_cacheSize = 16;
static int s_tessellationLevel = 9;
const static int c_maxTessellationLevel = 32;

static bool s_doBenchmark = false;
static float s_mts = 0.0f;


// Forward declarations.
void initPatchGeometry(int density);


// Mouse event handler.
static void mouse(int button, int state, int x, int y)
{
	ui.mouse(button, state, x, y);
}

// Mouse motion event handler.
static void motion(int x, int y)
{
	ui.mouseMotion(x, y);
}

// Key event handler.
static void keyboard(unsigned char c, int x, int y)
{
	ui.keyboard(c, x, y);

    if (c == 27 || c == 'q' || c == 'Q')
    {
        exit(0);
    }
    else if (c == 'w' || c == 'W')
    {
        s_drawWireframe = !s_drawWireframe;
    }
    else if (c == 'c' || c == 'C')
    {
        s_drawControlMesh = !s_drawControlMesh;
    }
    else if (c == '1')
    {
		s_drawMethod = DrawMethod_Tessellation;
    }
    else if (c == '2')
    {
		s_drawMethod = DrawMethod_Evaluators;
    }
    else if (c == '3')
    {
		s_drawMethod = DrawMethod_Static;
    }
	else if (c == '-')
    {
		if (--s_tessellationLevel < 1) s_tessellationLevel = 1;
		initPatchGeometry(s_tessellationLevel+1);
    }
	else if (c == '+' || c == '=')
    {
		if (++s_tessellationLevel > c_maxTessellationLevel) s_tessellationLevel = c_maxTessellationLevel;
		initPatchGeometry(s_tessellationLevel+1);
    }
	else if (c == ' ')
    {
		s_animate = !s_animate;
    }

    glutPostRedisplay();
}

static void drawTessellatedTeapot(int count = 1)
{
	glPushMatrix();

    if (s_drawMethod == DrawMethod_Tessellation)
    {
		// Apply the standard teapot transform.
		glRotated( 270.0, 1.0, 0.0, 0.0 );
		glScaled( 0.5, 0.5, 0.5 );
		glTranslated( 0.0, 0.0, -1.5 );

		// Select program.
		glUseProgram(s_tessellationProgram);

		GLint colorUniform = glGetUniformLocation(s_tessellationProgram, "color");
		glUniform3f(colorUniform, 1.0f, 0.0f, 0.0f);

		glEnableClientState(GL_VERTEX_ARRAY);
#ifdef USE_STRIPS
		glPrimitiveRestartIndexNV(0xFFFF);
		glEnableClientState(GL_PRIMITIVE_RESTART_NV);
#endif
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, s_vbHandle);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, s_ibHandle);
		glVertexPointer(2, GL_FLOAT, 0, 0);

		for (int i = 0; i < count; i++)
		{
			// Draw patches.
#ifdef USE_STRIPS
			glDrawElementsInstancedEXT(GL_TRIANGLE_STRIP, s_patchIndexCount, GL_UNSIGNED_SHORT, 0, c_teapotPatchCount);
#else
			glDrawElementsInstancedEXT(GL_TRIANGLES, s_patchIndexCount, GL_UNSIGNED_SHORT, 0, c_teapotPatchCount);
#endif
		}

		glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
		glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, 0);
		glDisableClientState(GL_VERTEX_ARRAY);

		glUseProgram(0);
	}
	else
	{
		glUseProgram(s_standardProgram);

		GLint colorUniform = glGetUniformLocation(s_standardProgram, "color");
		glUniform3f(colorUniform, 1.0f, 0.0f, 0.0f);

		if (s_drawMethod == DrawMethod_Evaluators)
		{
			// Apply the standard teapot transform.
			glRotated( 270.0, 1.0, 0.0, 0.0 );
			glScaled( 0.5, 0.5, 0.5 );
			glTranslated( 0.0, 0.0, -1.5 );

			// Mirror teapot.
			glScaled( 1, -1, 1 );

			glEnable(GL_AUTO_NORMAL);
			glEnable(GL_MAP2_VERTEX_3);
			//glEnable( GL_MAP2_TEXTURE_COORD_2 );

			//static float tex[4][2] = {{0,0}, {1,0}, {0,1}, {1,1}};

			for (int i = 0; i < count; i++)
			{
				// Draw teapot using evaluators.
				for (int p = 0; p < c_teapotPatchCount; p++)
				{
					GLfloat patch[16][3];
					for (int j = 0; j < 16; j++)
					{
						patch[j][0] = teapotVertices[3 * (teapotIndices[16 * p + j] - 1) + 0];
						patch[j][1] = teapotVertices[3 * (teapotIndices[16 * p + j] - 1) + 1];
						patch[j][2] = teapotVertices[3 * (teapotIndices[16 * p + j] - 1) + 2];
					}

					//glMap2d(GL_MAP2_TEXTURE_COORD_2, 0.0, 1.0, 2, 2, 0.0, 1.0, 4, 2, &tex[0][0]);
					glMap2f(GL_MAP2_VERTEX_3, 0.0, 1.0, 3, 4, 0.0, 1.0, 12, 4, &patch[0][0]);
					glMapGrid2d(s_tessellationLevel+1, 0.0, 1.0, s_tessellationLevel+1, 0.0, 1.0);
					glEvalMesh2(GL_FILL, 0, s_tessellationLevel+1, 0, s_tessellationLevel+1);
				}
			}

			glDisable(GL_AUTO_NORMAL);
			glDisable(GL_MAP2_VERTEX_3);
		}
		else
		{
			glScaled( 0.5, 0.5, 0.5 );
			glTranslated( 0.0, -1.5, 0.0 );

			glBindBuffer(GL_ARRAY_BUFFER, s_vertexBuffer);
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_indexBuffer);

			glVertexPointer(model.getPositionSize(), GL_FLOAT, model.getCompiledVertexSize() * sizeof(float), (void *)(model.getCompiledPositionOffset() * sizeof(float)));
			glNormalPointer(GL_FLOAT, model.getCompiledVertexSize() * sizeof(float), (void *)(model.getCompiledNormalOffset() * sizeof(float)));
			//glTexCoordPointer(model.getTexCoordSize(), GL_FLOAT, model.getCompiledVertexSize() * sizeof(float), (void *)(model.getCompiledPositionOffset() * sizeof(float)));

			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_NORMAL_ARRAY);

			for (int i = 0; i < count; i++)
			{
				// Draw reference teapot using display list.
				glDrawElements(GL_TRIANGLES, s_indexCount, GL_UNSIGNED_INT, 0);
			}

			glDisableClientState(GL_VERTEX_ARRAY);
			glDisableClientState(GL_NORMAL_ARRAY);

			glBindBuffer(GL_ARRAY_BUFFER, 0);
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		}

		glUseProgram(0);
    }

	glPopMatrix();
}

static float benchmark(int count)
{
#ifdef USE_GL_TIMER
    GLuint query;
    GLuint64EXT timeElapsed = 0;
    glGenQueries(1, &query);
    glBeginQuery(GL_TIME_ELAPSED_EXT, query);
#else
    glFinish();
    int starttime = glutGet(GLUT_ELAPSED_TIME);
#endif

	drawTessellatedTeapot(count);

#ifdef USE_GL_TIMER
    glEndQuery(GL_TIME_ELAPSED_EXT);

    glGetQueryObjectui64vEXT(query, GL_QUERY_RESULT, &timeElapsed);
    glDeleteQueries(1, &query);
    float time = timeElapsed / (1e9);    // Nanosecons!
#else
    glFinish();
    int endtime = glutGet(GLUT_ELAPSED_TIME);
    float time = (endtime - starttime) / 1000.0;
#endif

    float mts = count / (time*1e6);

	if (s_drawMethod == DrawMethod_Static)
	{
		mts *= s_indexCount / 4;
	}
	else
	{
		mts *= c_teapotPatchCount * (s_tessellationLevel+1) * (s_tessellationLevel+1) * 2;
	}

    printf("%f secs, %f Mtriangles/sec\n", time/count, mts);

	return mts;
}

// Do UI.
static void doUI()
{
    static nv::Rect none;
    
    ui.begin();

    ui.beginGroup(nv::GroupFlags_GrowDownFromLeft);

        ui.doCheckButton(none, "Draw control mesh", &s_drawControlMesh);

		ui.doCheckButton(none, "Draw wireframe", &s_drawWireframe);

        ui.beginGroup(nv::GroupFlags_GrowRightFromBottom|nv::GroupFlags_LayoutNoMargin);
		{
			ui.doLabel(none, "Rendering method:");

			const char * methods[] = { "Instanced Tessellation", "OpenGL Evaluators", "Static Mesh" };
			ui.doComboBox(none, 3, methods, &s_drawMethod);
		}
		ui.endGroup();

		if (s_drawMethod == DrawMethod_Tessellation)
		{
			if (ui.doCheckButton(none, "Optimize vertex cache", &s_preWarmCache))
			{
				initPatchGeometry(s_tessellationLevel+1);
			}
		}

    ui.endGroup();

	if (s_drawMethod != DrawMethod_Static)
    {
        ui.beginGroup(nv::GroupFlags_GrowRightFromBottom|nv::GroupFlags_LayoutNoMargin);
        
		ui.doLabel(none, "Tessellation Level:");

		nv::Rect rect(0, 0, s_width - 150, 0);

		float level = s_tessellationLevel;
        if (ui.doHorizontalSlider(rect, 1.0f, c_maxTessellationLevel, &level))
		{
			if (s_tessellationLevel != (int)level)
			{
				s_tessellationLevel = level;
				initPatchGeometry(s_tessellationLevel+1);
			}
		}

		char levelText[64];
		sprintf(levelText, "%i", s_tessellationLevel);
		ui.doLabel(none, levelText);

        ui.endGroup();
    }


    ui.beginGroup(nv::GroupFlags_GrowDownFromRight);

        if (ui.doButton(none, "Benchmark"))
        {
			s_doBenchmark = true;
        }

        if (s_mts != 0.0f)
        {
            char text[256];
            sprintf(text, "%.2f Mtriangles/sec", s_mts);
            ui.doLabel(none, text);
        }

    ui.endGroup();

    // Pass non-ui mouse events to the manipulator
	ui.updateManipulator(s_trackball);

    ui.end();
}

// Display event handler.
static void display()
{
	if (s_doBenchmark)
	{
		s_doBenchmark = false;
		s_mts = benchmark(128);
	}

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(20.0f, (double)s_width/s_height, 0.25f, 5.0f);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    s_trackball.applyTransform();

    if (s_drawWireframe)
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    }
    else
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    }

    glScalef(0.2, 0.2, 0.2);
    glColor3f(1, 1, 1);

	drawTessellatedTeapot();

    if (s_drawControlMesh)
    {
        glPushMatrix();

        // Apply the standard teapot transform.
        glRotated( 270.0, 1.0, 0.0, 0.0 );
        glScaled( 0.5, 0.5, 0.5 );
        glTranslated( 0.0, 0.0, -1.5 );

        glColor3f(0,0,1);

        // Draw control mesh using inmediate mode.
        glBegin(GL_LINES);

        for(int p = 0; p < c_teapotPatchCount; p++)
        {
            int v[16];
            for(int i = 0; i < 16; i++)
            {
                v[i] = 3 * (teapotIndices[p * 16 + i] - 1);
            }

            for(int i = 0; i < 4; i++)
            {
                glVertex3fv(teapotVertices + v[4*i+0]);
                glVertex3fv(teapotVertices + v[4*i+1]);

                glVertex3fv(teapotVertices + v[4*i+1]);
                glVertex3fv(teapotVertices + v[4*i+2]);

                glVertex3fv(teapotVertices + v[4*i+2]);
                glVertex3fv(teapotVertices + v[4*i+3]);

                glVertex3fv(teapotVertices + v[i+0]);
                glVertex3fv(teapotVertices + v[i+4]);

                glVertex3fv(teapotVertices + v[i+4]);
                glVertex3fv(teapotVertices + v[i+8]);

                glVertex3fv(teapotVertices + v[i+8]);
                glVertex3fv(teapotVertices + v[i+12]);
            }
        }

        glEnd();

        glPopMatrix();
    }

	doUI();

    glutSwapBuffers();
}

// Reshape event handler.
static void reshape(int w, int h)
{
	ui.reshape(w, h);

    s_width = w;
    s_height = h;
    glViewport(0, 0, s_width, s_height);
    s_trackball.reshape(w, h);
}

// Idle event handler.
void idle()
{
    if (s_animate) {
        s_trackball.idle();
    }
    glutPostRedisplay();
}

// Main menu event handler.
static void mainMenu(int i)
{
	keyboard((unsigned char) i, 0, 0);
}


// Compile shaders and link program.
static bool initShaders()
{
    char * vertexShaderText = new char[strlen(s_tessellationVertexShaderText) + 1024];
	sprintf(vertexShaderText, s_tessellationVertexShaderText, c_teapotIndexCount);

    GLuint tessellationVertexShader = nv::CompileGLSLShader(GL_VERTEX_SHADER, vertexShaderText);
    if (tessellationVertexShader == 0) 
    {
        printf("Vertex shader failed to compile\n");
        return false;
    }

    delete [] vertexShaderText;

    GLuint standardVertexShader = nv::CompileGLSLShader(GL_VERTEX_SHADER, s_standardVertexShaderText);
    if (standardVertexShader == 0) 
    {
        printf("Vertex shader failed to compile\n");
        return false;
    }

    GLuint fragmentShader = nv::CompileGLSLShader(GL_FRAGMENT_SHADER, s_fragmentShaderText);
    if (fragmentShader == 0) 
    {
        printf("Fragment shader failed to compile\n");
        return false;
    }

    s_tessellationProgram = nv::LinkGLSLProgram(tessellationVertexShader, fragmentShader);
    if (s_tessellationProgram == 0) 
    {
        printf("Program failed to link\n");
        return false;
    }

    s_standardProgram = nv::LinkGLSLProgram(standardVertexShader, fragmentShader);
    if (s_standardProgram == 0) 
    {
        printf("Program failed to link\n");
        return false;
    }

    return true;
}


// Allocate bindable uniform buffers for teapot control mesh.
void initUniformBuffers()
{
    int maxVertexBindableUniforms;
    glGetIntegerv(GL_MAX_VERTEX_BINDABLE_UNIFORMS_EXT, &maxVertexBindableUniforms);
    assert(maxVertexBindableUniforms >= 2);

    int maxBindableUniformSize;
    glGetIntegerv(GL_MAX_BINDABLE_UNIFORM_SIZE_EXT, &maxBindableUniformSize);
    assert(4 * sizeof(float) * c_teapotVertexCount <= maxBindableUniformSize);
    assert(sizeof(teapotIndices) <= maxBindableUniformSize);

    // create vertex buffer.
    GLint vlocation = glGetUniformLocation(s_tessellationProgram, "vertices");
	GLint vsize = glGetUniformBufferSizeEXT(s_tessellationProgram, vlocation);
	assert(vsize == 4 * sizeof(float) * c_teapotIndexCount);

    // In the present implementation all constants must be 128-bit aligned, so an array of 
    // vec3s is padded to vec4s. A future OpenGL extension may remove this limitation.
	float * tmp = new float[c_teapotIndexCount * 4];
    for(int i = 0; i < c_teapotIndexCount; i++)
    {
		int idx = teapotIndices[i] - 1;
        tmp[4 * i + 0] = teapotVertices[3 * idx + 0];
        tmp[4 * i + 1] = teapotVertices[3 * idx + 1];
        tmp[4 * i + 2] = teapotVertices[3 * idx + 2];
        tmp[4 * i + 3] = 1.0f;    // pad
    }

    GLuint vbuffer;
    glGenBuffers(1, &vbuffer);
    glBindBuffer(GL_UNIFORM_BUFFER_EXT, vbuffer);
    glBufferData(GL_UNIFORM_BUFFER_EXT, vsize, tmp, GL_STATIC_READ);
    glUniformBufferEXT(s_tessellationProgram, vlocation, vbuffer);

    delete [] tmp;
}


// Allocate vertex and index buffer for rectangular patch.
// Optimize vertex and index buffer for perfect cache coherency.
static void initPatchGeometry(int density)
{
	if (s_vbHandle) glDeleteBuffersARB(1, &s_vbHandle);
	if (s_ibHandle) glDeleteBuffersARB(1, &s_ibHandle);

	// @@ Use multiple strips when density >= cache size.

#ifdef USE_STRIPS
	s_patchIndexCount = density * (2 * density + 3);
	if (s_preWarmCache)
	{
		//s_patchIndexCount += 2 * density + 2;
		s_patchIndexCount += 1 * density + 3;
	}
#else
	s_patchIndexCount = density * density * 6;

	int strips = 0;
	if (s_preWarmCache)
	{
		s_patchIndexCount += 3 * density;

		/*while (density + 1 > (strips + 1) * s_cacheSize)
		{
			strips++;
		}*/
	}
#endif
    s_patchVertexCount = (density + 1) * (density + 1);

    float * const vertexPtr = new float[2 * s_patchVertexCount];
	GLushort * const indexPtr = new GLushort[s_patchIndexCount];

    int idx = 0;
    for(int i = 0; i <= density; i++)
    {
        float v = float(i) / density;
            
        for(int e = 0; e <= density; e++)
        {
            float u = float(e) / density;

            vertexPtr[idx++] = u;
            vertexPtr[idx++] = v;
        }
    }
    assert(idx == 2 * s_patchVertexCount);

    idx = 0;

#ifdef USE_STRIPS
	if (s_preWarmCache)
	{
		// Create degenerate triangles to prewarm cache.
		indexPtr[idx++] = i;
		indexPtr[idx++] = i + 1;
		
		for(int i = 0; i < density; i++)
		{
			indexPtr[idx++] = i + 1;
		//	indexPtr[idx++] = i + 1;
		}

		indexPtr[idx++] = 0xFFFF; // Restart strip.
	}

	// Real triangles:
    for(int i = 0; i < density; i++)
    {
        indexPtr[idx++] = (density+1) * (i+1);
        indexPtr[idx++] = (density+1) * (i+0);

        for(int e = 0; e < density; e++)
        {
            indexPtr[idx++] = (density+1) * (i+1) + (e+1);
            indexPtr[idx++] = (density+1) * (i+0) + (e+1);
        }

        indexPtr[idx++] = 0xFFFF; // Restart strip.
    }
    assert(idx == s_patchIndexCount);
#else
	if (s_preWarmCache)
	{
		// Create degenerate triangles to prewarm cache.
		for(int i = 0; i < density; i++)
		{
			indexPtr[idx++] = i;
			indexPtr[idx++] = i + 1;
			indexPtr[idx++] = i + 1;
		}
	}

	// Real triangles:
    for(int i = 0; i < density; i++)
    {
        for(int e = 0; e < density; e++)
        {
            indexPtr[idx++] = (density+1) * (i+0) + (e+0);
            indexPtr[idx++] = (density+1) * (i+0) + (e+1);
            indexPtr[idx++] = (density+1) * (i+1) + (e+0);

            indexPtr[idx++] = (density+1) * (i+1) + (e+0);
            indexPtr[idx++] = (density+1) * (i+0) + (e+1);
            indexPtr[idx++] = (density+1) * (i+1) + (e+1);
        }
    }
    assert(idx == s_patchIndexCount);
#endif

	// @@ Reorder vertices according to indices to maximize cache coherency.

    glGenBuffersARB(1, &s_vbHandle);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, s_vbHandle);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, s_patchVertexCount * sizeof(float) * 2, vertexPtr, GL_STATIC_DRAW_ARB);

    glGenBuffersARB(1, &s_ibHandle);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER, s_ibHandle);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER, s_patchIndexCount * sizeof(short), indexPtr, GL_STATIC_DRAW_ARB);

    delete [] vertexPtr;
    delete [] indexPtr;
}

bool initRawGeometry()
{
	static nv::SDKPath sdkPath;

    std::string resolved_path;
    if (!sdkPath.getFilePath("../media/models/teapot.obj", resolved_path))
	{
		printf("Cannot find 'teapot.obj'\n");
		return false;
	}

	if (!model.loadModelFromFile(resolved_path.c_str()))
	{
		printf("Error loading 'teapot.obj'\n");
		return false;
	}

	model.computeNormals();
	model.compileModel();

	glGenBuffers(1, &s_indexBuffer);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, s_indexBuffer);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, model.getCompiledIndexCount() * sizeof(int), model.getCompiledIndices(), GL_STATIC_READ);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

	s_indexCount = model.getCompiledIndexCount();

	glGenBuffers(1, &s_vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, s_vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, model.getCompiledVertexCount() * model.getCompiledVertexSize() * sizeof(float), model.getCompiledVertices(), GL_STATIC_READ);

	glBindBuffer(GL_ARRAY_BUFFER, 0);

	return true;
}


// Program entrypoint.
int main(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_ALPHA | GLUT_DEPTH | GLUT_STENCIL);
    glutInitWindowSize(s_width, s_height);
    glutCreateWindow("Instanced tessellation example");

    glutDisplayFunc(display);
    glutKeyboardFunc(keyboard);
    glutReshapeFunc(reshape);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);

    s_trackball.setDollyActivate(GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    s_trackball.setPanActivate(GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    s_trackball.setDollyPosition(-2.0f);

    printf("This example draws a mesh composed of bezier patches using intancing to emulate tessellation.\n\n");
    printf("Use the following keys to toggle options:\n");
    printf("[1] Draw using instanced tessellation.\n");
    printf("[2] Draw using OpenGL evaluators.\n");
	printf("[3] Draw static mesh.\n");
	printf("[c] Toggle drawing of the control mesh.\n");
    printf("[w] Toggle wireframe rendering.\n");
	printf("[+/-] Increase/Decrease tessellation level.\n");
    printf("[q] Exit the application.\n\n");
    printf("Use the mouse to manipulate camera:\n");
    printf("Left mouse button to rotate around the object.\n");
    printf("Left mouse button + Shift to dolly.\n");
    printf("Left mouse button + Ctrl to pan.\n\n");

    glewInit();
    if (!GLEW_VERSION_2_0 || !GLEW_ARB_vertex_buffer_object || !GLEW_EXT_gpu_shader4 || !GLEW_EXT_bindable_uniform || !GLEW_EXT_draw_instanced)
    {
        printf( "Required extensions not supported.\n");
        printf( "This sample requires:\n");
        printf( "  OpenGL version 2.0 or later\n");
        printf( "  GL_ARB_vertex_buffer_object\n");
        printf( "  GL_EXT_gpu_shader4\n");
        printf( "  GL_EXT_bindable_uniform\n");
        printf( "  GL_EXT_draw_instanced\n");
        return 1;
    }

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	const char * renderer = (const char *)glGetString(GL_RENDERER);
	printf("Renderer: %s\n", renderer);
	printf("Version: %s\n", glGetString(GL_VERSION));

	// These are effective cache sizes obtained experimentally.
	if (strncmp("GeForce 8800", renderer, 12) == 0)
	{
		s_cacheSize = 21;
	}
	else if (strncmp("GeForce 8600", renderer, 12) == 0)
	{
		s_cacheSize = 52;
	}

	printf("Cache Size: %d\n", s_cacheSize);

     if (!initShaders())
    {
        return 1;
    }

    initUniformBuffers();
    initPatchGeometry(s_tessellationLevel+1);
	if (!initRawGeometry())
	{
		return 1;
	}

    // Set GL state.
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_CULL_FACE);

    glutMainLoop();

    return 0;
}